/*
 * This module implements framebuffer driver for LCD Wintek WG19232.
 */

#include <linux/module.h>// main header
#include <linux/cdev.h>// cdev_alloc
#include <linux/init.h>// __init
#include <linux/kernel.h>// printk()
#include <linux/types.h>// size_t
#include <linux/slab.h>// kmalloc()
#include <linux/delay.h>// udelay
#include <linux/types.h>// u32
#include <linux/fb.h>// frame buffer
#include <linux/kthread.h>
//#include <linux/gpio.h>
#include <mach/gpio.h>// at91 specific
#include <mach/at91_pio.h>// at91 specific

#define DISPLAY_SHOW_SPLASH 1
#define SPLASH_X_POS 0
#define SPLASH_Y_POS 0

#define DEBUG 0

#if DEBUG
# define DPRINT(args...) printk(args)
#else
# define DPRINT(args...)
#endif

#define xDPRINT(args...)

#if DISPLAY_SHOW_SPLASH
#include "splash.xbm"
#endif

#define REFRESH_PERIOD	(HZ / 5)// 5 fps

#define BASE_PIO		(AT91_BASE_SYS + AT91_PIOB)//0xFFFFF600//

#define SODR			(PIO_SODR / 4)// we are using u32* piob
#define CODR			(PIO_CODR / 4)// we are using u32* piob

#define LCD_BACKLIGHT	AT91_PIN_PB11
#define LCD_RS			AT91_PIN_PB22
//#define LCD_BIT_RS		22//((LCD_RS - PIN_BASE) % 32)
#define LCD_E			AT91_PIN_PB23
//#define LCD_BIT_E		23//((LCD_E - PIN_BASE) % 32)
//#define LCD_DATA_SHIFT	24//((LCD_D0 - PIN_BASE) % 32)
#define LCD_D0			AT91_PIN_PB24
#define LCD_D1			AT91_PIN_PB25
#define LCD_D2			AT91_PIN_PB26
#define LCD_D3			AT91_PIN_PB27
#define LCD_D4			AT91_PIN_PB28
#define LCD_D5			AT91_PIN_PB29
#define LCD_D6			AT91_PIN_PB30
#define LCD_D7			AT91_PIN_PB31

#define LCD_WIDTH		192
#define LCD_HEIGHT		32// <= 255, see lcd_show
#define LCD_MEM_SIZE    (LCD_WIDTH * LCD_HEIGHT / 8)

static struct fb_info wg19232_fb_info;
static u8* screen_shadow;

//static volatile u32 *piob;// io base address to send data to lcd
static volatile u32 *pio_sodr;// io sodr address to send data to lcd
static volatile u32 *pio_codr;// io codr address to send data to lcd
struct task_struct *wg19232_fb_thread_task;

static struct fb_fix_screeninfo wg19232_fb_fix __initdata = {
	.id				= "wg19232fb",
	.type			= FB_TYPE_PACKED_PIXELS,
	.visual			= FB_VISUAL_MONO01,// Monochr. 1=Black 0=White
	.xpanstep		= 0,
	.ypanstep		= 0,
	.ywrapstep		= 0,
	.line_length	= (LCD_WIDTH - 1 + 8) / 8,
	.accel			= FB_ACCEL_NONE,
};

static struct fb_var_screeninfo wg19232_fb_default __initdata = {
	.xres			= LCD_WIDTH,
	.yres			= LCD_HEIGHT,
	.xres_virtual	= LCD_WIDTH,
	.yres_virtual	= LCD_HEIGHT,
	.bits_per_pixel	= 1,
	//.red			= { 0, 8, 0 },
	//.green		= { 0, 8, 0 },
	//.blue			= { 0, 8, 0 },
	.activate		= FB_ACTIVATE_TEST,
	.height			= LCD_HEIGHT,//-1,
	.width			= LCD_WIDTH,//-1,
	.left_margin	= 0,
	.right_margin	= 0,
	.upper_margin	= 0,
	.lower_margin	= 0,
	.vmode			= FB_VMODE_NONINTERLACED,
	.nonstd			= 0,
};
/*
static void wg19232_data(bool rs, u8 data, unsigned long wait) 
{
	piob[CODR] = 0xFF << ((LCD_D0 - PIN_BASE) % 32);//LCD_DATA_SHIFT;// Data clear
	piob[SODR] = data << ((LCD_D0 - PIN_BASE) % 32);//LCD_DATA_SHIFT;// Data set

	piob[rs ? SODR : CODR] = 1 << ((LCD_RS - PIN_BASE) % 32);//LCD_BIT_RS;// RS

	piob[SODR] = 1 << ((LCD_E - PIN_BASE) % 32);//LCD_BIT_E;// Set E
	udelay( 100 );				// Bit E must be set > 120mks
	piob[CODR] = 1 << ((LCD_E - PIN_BASE) % 32);//LCD_BIT_E;// Clear E

	udelay( wait );				// Data must be set while set RS
}
*/
static void wg19232_data(bool rs, u8 data, unsigned long wait) 
{
	volatile u32* p = pio_codr;
	*pio_codr = 0xFF << ((LCD_D0 - PIN_BASE) % 32);//LCD_DATA_SHIFT;// Data clear
	*pio_sodr = data << ((LCD_D0 - PIN_BASE) % 32);//LCD_DATA_SHIFT;// Data set

	if( rs ) p = pio_sodr;
	*p = 1 << ((LCD_RS - PIN_BASE) % 32);//LCD_BIT_RS;// RS

	*pio_sodr = 1 << ((LCD_E - PIN_BASE) % 32);//LCD_BIT_E;// Set E
	udelay( 100 );				// Bit E must be set > 120mks
	*pio_codr = 1 << ((LCD_E - PIN_BASE) % 32);//LCD_BIT_E;// Clear E

	udelay( wait );				// Data must be set while set RS
}

static void wg19232_show(void) 
{
	unsigned int x;
	u8 y;
	unsigned int i = 0;

	for( y = 0; y < LCD_HEIGHT; y++ )
	{
		wg19232_data( false, (1<<7) | y, 9 );
		wg19232_data( false, (1<<7), 9 );

		for( x = 0; x < LCD_WIDTH / 8; x++ )
			wg19232_data( true, screen_shadow[i++], 10 );
	}
}

static int wg19232_fb_thread(void *param)
{
	while( !kthread_should_stop() )
	{
		if( memcmp(screen_shadow, wg19232_fb_info.screen_base, LCD_MEM_SIZE) )
		{
			memcpy( screen_shadow, wg19232_fb_info.screen_base, LCD_MEM_SIZE );
			wg19232_show();
		}
		schedule_timeout_interruptible( REFRESH_PERIOD );
	}

	return 0;
}

static int wg19232_fb_blank(int blank_mode, struct fb_info *info)
{
	switch (blank_mode)
	{
	case FB_BLANK_POWERDOWN:
	case FB_BLANK_VSYNC_SUSPEND:
	case FB_BLANK_HSYNC_SUSPEND:
	case FB_BLANK_NORMAL:
		gpio_set_value( LCD_BACKLIGHT, 0 );
		break;
	case FB_BLANK_UNBLANK:
		gpio_set_value( LCD_BACKLIGHT, 1 );
		break;
	}
	return 0;
}

static struct fb_ops wg19232_fops =
{
	.owner			= THIS_MODULE,
	.fb_blank		= wg19232_fb_blank,
	//.fb_fillrect	= sys_fillrect,
	//.fb_copyarea	= sys_copyarea,
	//.fb_imageblit	= sys_imageblit
};

#if DISPLAY_SHOW_SPLASH
static void __init show_splash(void)
{
	u8* buff = (u8*)wg19232_fb_info.screen_base;
	u32 sx, sy, x, y;
	u8 d = 0;
	sy = 0;
	for( y = SPLASH_Y_POS * wg19232_fb_fix.line_length;
		y < wg19232_fb_fix.line_length * (((splash_height + SPLASH_Y_POS) < LCD_HEIGHT) ?
									(splash_height + SPLASH_Y_POS) : LCD_HEIGHT);
		y += wg19232_fb_fix.line_length )
	{
		sx = 0;
		for( x = SPLASH_X_POS; x < (((splash_width + SPLASH_X_POS) < LCD_WIDTH) ? (splash_width + SPLASH_X_POS) : LCD_WIDTH); x++)
		{
			if( !(sx & 7) ) d = splash_bits[sy + (sx >> 3)];
			if( d & (1 << (sx & 7)) )
			{
				buff[y + (x >> 3)] |= (0x80 >> (x & 7));
			}
			sx++;
		}
		sy += (splash_width - 1 + 8) / 8;
	}
}
#endif
/*
this is do not compiling
static struct gpio wg19232_gpios[] = {
	//unsigned gpio; unsigned long flags; const char *label;
	{ LCD_BACKLIGHT,	GPIOF_OUT_INIT_HIGH,	"LCD backlight" },
	{ LCD_RS,			GPIOF_OUT_INIT_LOW,		"LCD rs" },
	{ LCD_E,			GPIOF_OUT_INIT_LOW,		"LCD e" },
	{ LCD_D0,			GPIOF_OUT_INIT_LOW,		"LCD d0" },
	{ LCD_D1,			GPIOF_OUT_INIT_LOW,		"LCD d1" },
	{ LCD_D2,			GPIOF_OUT_INIT_LOW,		"LCD d2" },
	{ LCD_D3,			GPIOF_OUT_INIT_LOW,		"LCD d3" },
	{ LCD_D4,			GPIOF_OUT_INIT_LOW,		"LCD d4" },
	{ LCD_D5,			GPIOF_OUT_INIT_LOW,		"LCD d5" },
	{ LCD_D6,			GPIOF_OUT_INIT_LOW,		"LCD d6" },
	{ LCD_D7,			GPIOF_OUT_INIT_LOW,		"LCD d7" }
};
*/
u8 wg19232_initdata[] __initdata =
{
	(1<<5) | (1<<4),
	(1<<5) | (1<<4),
	(1<<3) | (1<<2),
	(1<<0),
	(1<<2) | (1<<1) | (1<<0),
	(1<<5) | (1<<4) | (1<<2),
	(1<<1),
	(1<<5) | (1<<4) | (1<<2) | (1<<1)
};

static int __init wg19232_init(void)
{
	int err = -ENOMEM;
	int i;
	//
	screen_shadow = alloc_pages_exact( LCD_MEM_SIZE, GFP_DMA | __GFP_ZERO );//(u8*)kzalloc( LCD_MEM_SIZE, GFP_KERNEL );
	if( !screen_shadow )
	{
		printk("cannot alloc memory for shadow buffer\n");
		return -ENOMEM;
	}
	wg19232_fb_info.screen_base = alloc_pages_exact( LCD_MEM_SIZE, GFP_DMA | __GFP_ZERO );//kzalloc( LCD_MEM_SIZE, GFP_KERNEL );
	if( !wg19232_fb_info.screen_base )
	{
		printk("cannot alloc memory for screen buffer\r\n");
		goto exit_free_shadow;
	}
	wg19232_fb_fix.smem_start = (unsigned long)virt_to_phys( wg19232_fb_info.screen_base );
	wg19232_fb_fix.smem_len = LCD_MEM_SIZE;
	wg19232_fb_info.fbops = &wg19232_fops;
	wg19232_fb_info.var = wg19232_fb_default;
	wg19232_fb_info.fix = wg19232_fb_fix;
	wg19232_fb_info.flags = FBINFO_FLAG_DEFAULT;
	/*
	piob = (u32*)ioremap_nocache( BASE_PIO, 64 );
	if( !piob )
	{
		printk("unable remap GPIO\r\n");
		goto exit_free_screen;
	}
	*/
	pio_sodr = (u32*)ioremap_nocache( BASE_PIO + PIO_SODR, 4 );
	if( !pio_sodr )
	{
		printk("unable remap GPIO\r\n");
		goto exit_free_screen;
	}
	pio_codr = (u32*)ioremap_nocache( BASE_PIO + PIO_CODR, 4 );
	if( !pio_codr )
	{
		printk("unable remap GPIO\r\n");
		goto exit_free_sodr;
	}
	/* this is do not compiling
	err = gpio_request_array( wg19232_gpios, ARRAY_SIZE(wg19232_gpios) );
	if( err )
	{
		printk("unable to init GPIO for output\r\n");
		goto exit_ioremap;
	}
	*/
	err = gpio_request(LCD_BACKLIGHT, "lcd backlight");
	if( err || gpio_direction_output( LCD_BACKLIGHT, 1 ) ) { printk("unable to set GPIO %d for output\n", LCD_BACKLIGHT); goto exit_ioremap; }
	err = gpio_request(LCD_RS, "lcd rs");
	if( err || gpio_direction_output( LCD_RS, 0 ) ) { printk("unable to set GPIO %d for output\n", LCD_RS); goto exit_free_bl; }
	err = gpio_request(LCD_E, "lcd e");
	if( err || gpio_direction_output( LCD_E, 0 ) ) { printk("unable to set GPIO %d for output\n", LCD_E); goto exit_free_rs; }
	err = gpio_request(LCD_D0, "lcd d0");
	if( err || gpio_direction_output( LCD_D0, 0 ) ) { printk("unable to set GPIO %d for output\n", LCD_D0); goto exit_free_e; }
	err = gpio_request(LCD_D1, "lcd d1");
	if( err || gpio_direction_output( LCD_D1, 0 ) ) { printk("unable to set GPIO %d for output\n", LCD_D1); goto exit_free_d0; }
	err = gpio_request(LCD_D2, "lcd d2");
	if( err || gpio_direction_output( LCD_D2, 0 ) ) { printk("unable to set GPIO %d for output\n", LCD_D2); goto exit_free_d1; }
	err = gpio_request(LCD_D3, "lcd d3");
	if( err || gpio_direction_output( LCD_D3, 0 ) ) { printk("unable to set GPIO %d for output\n", LCD_D3); goto exit_free_d2; }
	err = gpio_request(LCD_D4, "lcd d4");
	if( err || gpio_direction_output( LCD_D4, 0 ) ) { printk("unable to set GPIO %d for output\n", LCD_D4); goto exit_free_d3; }
	err = gpio_request(LCD_D5, "lcd d5");
	if( err || gpio_direction_output( LCD_D5, 0 ) ) { printk("unable to set GPIO %d for output\n", LCD_D5); goto exit_free_d4; }
	err = gpio_request(LCD_D6, "lcd d6");
	if( err || gpio_direction_output( LCD_D6, 0 ) ) { printk("unable to set GPIO %d for output\n", LCD_D6); goto exit_free_d5; }
	err = gpio_request(LCD_D7, "lcd d7");
	if( err || gpio_direction_output( LCD_D7, 0 ) ) { printk("unable to set GPIO %d for output\n", LCD_D7); goto exit_free_d6; }
	//
	for( i = 0; i < sizeof(wg19232_initdata); i++ )
	{
		wg19232_data( false, wg19232_initdata[i], 200 );
	}

#if DISPLAY_SHOW_SPLASH
	show_splash();
#endif
	wg19232_fb_thread_task = kthread_run( wg19232_fb_thread, NULL, "wg19232fb" );
	if( IS_ERR(wg19232_fb_thread_task) )
	{
		printk("unable to run showing thread\r\n");
		goto exit_free_gpio;
	}
	//
	if( register_framebuffer(&wg19232_fb_info) < 0 )
	{
		printk("cannot register fb device\r\n");
		err = -EINVAL;
		goto exit_free_thread;
	}
	return err;

exit_free_thread:
	kthread_stop( wg19232_fb_thread_task );
//
exit_free_gpio:
	gpio_free( LCD_D7 );
exit_free_d6:
	gpio_free( LCD_D6 );
exit_free_d5:
	gpio_free( LCD_D5 );
exit_free_d4:
	gpio_free( LCD_D4 );
exit_free_d3:
	gpio_free( LCD_D3 );
exit_free_d2:
	gpio_free( LCD_D2 );
exit_free_d1:
	gpio_free( LCD_D1 );
exit_free_d0:
	gpio_free( LCD_D0 );
exit_free_e:
	gpio_free( LCD_E );
exit_free_rs:
	gpio_free( LCD_RS );
exit_free_bl:
	gpio_free( LCD_BACKLIGHT );
//exit_free_gpio:
	//gpio_free_array( wg19232_gpios, ARRAY_SIZE(wg19232_gpios) );
exit_ioremap:
	//iounmap( piob );
	iounmap( pio_codr );
exit_free_sodr:
	iounmap( pio_sodr );
exit_free_screen:
	free_pages_exact( wg19232_fb_info.screen_base, LCD_MEM_SIZE );//kfree( fb_info.screen_base );
exit_free_shadow:
	free_pages_exact( screen_shadow, LCD_MEM_SIZE );//kfree( screen_shadow );
	return err;
}

static void __exit wg19232_exit(void)
{
	unregister_framebuffer( &wg19232_fb_info );
	kthread_stop( wg19232_fb_thread_task );
	free_pages_exact( screen_shadow, LCD_MEM_SIZE );//kfree( screen_shadow );
	free_pages_exact( wg19232_fb_info.screen_base, LCD_MEM_SIZE );//kfree( fb_info.screen_base );
	gpio_set_value( LCD_BACKLIGHT, 0 );
	// this is do not compiling
	//gpio_free_array( wg19232_gpios, ARRAY_SIZE(wg19232_gpios) );
	//
	gpio_free( LCD_D7 );
	gpio_free( LCD_D6 );
	gpio_free( LCD_D5 );
	gpio_free( LCD_D4 );
	gpio_free( LCD_D3 );
	gpio_free( LCD_D2 );
	gpio_free( LCD_D1 );
	gpio_free( LCD_D0 );
	gpio_free( LCD_E );
	gpio_free( LCD_RS );
	gpio_free( LCD_BACKLIGHT );
	//
    //iounmap( piob );
	iounmap( pio_codr );
	iounmap( pio_sodr );
}

MODULE_DESCRIPTION("wg19232 GPIO frame buffer driver");
MODULE_LICENSE("GPL");

module_init( wg19232_init );
module_exit( wg19232_exit );
